# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
"""Add columns indicating corresponding numeric columns have NaN."""
import logging
import warnings
from typing import List, Optional

import numpy
import pandas as pd

from azureml.automl.core.shared.logging_utilities import function_debug_log_wrapped
from ...timeseries._time_series_data_set import TimeSeriesDataSet
from .._azureml_transformer import AzureMLTransformer

# Prevent warnings when using Jupyter
warnings.simplefilter("ignore")
pd.options.mode.chained_assignment = None


class MissingDummiesTransformer(AzureMLTransformer):
    """Add columns indicating corresponding numeric columns have NaN."""

    MARKER_VALUE_MISSING = 1
    MARKER_VALUE_NOT_MISSING = 0

    MISSING_DUMMIES_POSTFIX = "_WASNULL"

    def __init__(self, numerical_columns: List[str]) -> None:
        """
        Construct for MissingDummiesTransformer.

        :param numerical_columns: The columns that will be marked.
        :type numerical_columns: list
        :return:
        """
        super().__init__()
        self.numerical_columns = numerical_columns

    @function_debug_log_wrapped(logging.INFO)
    def fit(self, x: TimeSeriesDataSet, y: Optional[numpy.ndarray] = None) -> "MissingDummiesTransformer":
        """
        Fit function for MissingDummiesTransformer.

        :param x: Input data.
        :type x: azureml.automl.runtime._time_series_data_set.TimeSeriesDataSet
        :param y: Target values.
        :type y: numpy.ndarray
        :return: Class object itself.
        """
        return self

    @function_debug_log_wrapped(logging.INFO)
    def transform(self, x: TimeSeriesDataSet) -> TimeSeriesDataSet:
        """
        Transform function for MissingDummiesTransformer.

        :param x: Input data.
        :type x: azureml.automl.runtime._time_series_data_set.TimeSeriesDataSet
        :return: Result of MissingDummiesTransformer.
        """
        for col in self.numerical_columns:
            x.data[MissingDummiesTransformer.get_column_name(col)] = x.data[col].isnull().astype("int")
        return x

    @staticmethod
    def get_column_name(col: str) -> str:
        """
        Return the name of column marking nan in the given source column.

        :param col: the name of the source column.
        :return: the name of new feature column.
        """
        return col + MissingDummiesTransformer.MISSING_DUMMIES_POSTFIX

    @staticmethod
    def get_col_internal_type(col_name: str) -> Optional[str]:
        """
        Get the type of a column if it is generated by missing dummies transformer.

        :param col_name: The column name.
        :return: If column is generated by missing dummies transformer, return 'missing_dummies', else None.
        """
        if col_name.endswith(MissingDummiesTransformer.MISSING_DUMMIES_POSTFIX):
            return "missing_dummies"
        return None
